// ICON
//
// ---------------------------------------------------------------
// Copyright (C) 2004-2024, DWD, MPI-M, DKRZ, KIT, ETH, MeteoSwiss
// Contact information: icon-model.org
//
// See AUTHORS.TXT for a list of authors
// See LICENSES/ for license information
// SPDX-License-Identifier: BSD-3-Clause
// ---------------------------------------------------------------

#include "mo_lib_intp_rbf.hpp"
#include <Kokkos_Core.hpp>
#include <algorithm>
#include <gtest/gtest.h>
#include <numeric>
#include <vector>

// Free-function helpers for 3D and 4D array sizes (assumed column-major)
template <typename T> size_t num_elements_3d(int d1, int d2, int d3) {
  return static_cast<size_t>(d1) * d2 * d3;
}

template <typename T> size_t num_elements_4d(int d1, int d2, int d3, int d4) {
  return static_cast<size_t>(d1) * d2 * d3 * d4;
}

// Define a helper struct that holds the two types.
template <typename InT, typename OutT> struct MixedPrecision {
  using in_type = InT;
  using out_type = OutT;
};

// Define the list of type pairs we want to test.
typedef ::testing::Types<MixedPrecision<double, double>,
                         MixedPrecision<double, float>,
                         MixedPrecision<float, float>>
    MixedTypes;

class interp_dimensions {
public:
  // Constant dimensions.
  static constexpr int nproma = 3;          // inner loop length
  static constexpr int nlev = 4;            // number of vertical levels
  static constexpr int nblks_c = 2;         // number of cell blocks
  static constexpr int nblks_e = 2;         // number of edge blocks
  static constexpr int nblks_v = 2;         // number of vertex blocks
  static constexpr int rbf_c2grad_dim = 10; // fixed dimension
  static constexpr int rbf_vec_dim_c = 9;
  static constexpr int rbf_vec_dim_e = 4;

  // Parameter values.
  const int i_startblk = 0;
  const int i_endblk = 1; // Test blocks [0, 1]
  const int i_startidx_in = 0;
  const int i_endidx_in = nproma - 1;
  const int slev = 0;
  const int elev = nlev - 1;
  const bool lacc = false;      // Not using ACC-specific behavior.
  const bool acc_async = false; // No asynchronous execution.
};

// Define a typed test fixture for the functions which have the same input and
// output types
template <typename T>
class RbfInterpolTypedTestFixture : public ::testing::Test,
                                    public interp_dimensions {
public:
  // Data arrays.
  std::vector<T> p_cell_in;        // size: nproma * nlev * nblks_c
  std::vector<int> rbf_c2grad_idx; // size: rbf_c2grad_dim * nproma * nblks_c
  std::vector<int> rbf_c2grad_blk; // size: rbf_c2grad_dim * nproma * nblks_c
  std::vector<int> rbf_vec_idx_c;  // size: rbf_vec_dim_c * nproma * nblks_c
  std::vector<int> rbf_vec_blk_c;  // size: rbf_vec_dim_c * nproma * nblks_c
  std::vector<T>
      rbf_c2grad_coeff;  // size: rbf_c2grad_dim * 2 * nproma * nblks_c
  std::vector<T> grad_x; // size: nproma * nlev * nblks_c
  std::vector<T> grad_y; // size: nproma * nlev * nblks_c
  std::vector<T> p_vn_in;
  std::vector<T> rbf_vec_coeff_c;
  std::vector<T> p_u_out;
  std::vector<T> p_v_out;

  std::vector<int> rbf_vec_idx_e;
  std::vector<int> rbf_vec_blk_e;
  std::vector<T> rbf_vec_coeff_e;
  std::vector<T> p_vt_out;

  RbfInterpolTypedTestFixture() {
    size_t size3d = static_cast<size_t>(nproma) * nlev * nblks_c;
    size_t size3d_idx = static_cast<size_t>(rbf_c2grad_dim) * nproma * nblks_c;
    size_t size4d = static_cast<size_t>(rbf_c2grad_dim) * 2 * nproma * nblks_c;

    size_t size3d_vec_dim =
        static_cast<size_t>(rbf_vec_dim_c) * nproma * nblks_c;
    size_t size_4d_vec_dim =
        static_cast<size_t>(rbf_vec_dim_c) * 2 * nproma * nblks_c;

    size_t size3d_edge_lib =
        static_cast<size_t>(rbf_vec_dim_e) * nproma * nblks_c;
    size_t size_4d_edge_lib =
        static_cast<size_t>(rbf_vec_dim_e) * 2 * nproma * nblks_c;

    p_cell_in.resize(size3d, static_cast<T>(1));
    p_vn_in.resize(size3d, static_cast<T>(1));

    rbf_vec_idx_c.resize(size3d_vec_dim, 1);
    rbf_vec_blk_c.resize(size3d_vec_dim, 0);
    rbf_c2grad_idx.resize(size3d_idx, 1);
    rbf_c2grad_blk.resize(size3d_idx, 0); // Set block indices to 0 for testing.
    rbf_vec_idx_e.resize(size3d_vec_dim, 1);
    rbf_vec_blk_e.resize(size3d_vec_dim, 0);

    rbf_vec_coeff_c.resize(size_4d_vec_dim, static_cast<T>(1));
    rbf_c2grad_coeff.resize(size4d, static_cast<T>(1));
    rbf_vec_coeff_e.resize(size_4d_edge_lib, static_cast<T>(1));

    p_u_out.resize(size3d_vec_dim, static_cast<T>(0));
    p_v_out.resize(size3d_vec_dim, static_cast<T>(0));
    p_vt_out.resize(size3d_edge_lib, static_cast<T>(0));

    grad_x.resize(size3d, static_cast<T>(0));
    grad_y.resize(size3d, static_cast<T>(0));
  }
};

typedef ::testing::Types<float, double> MyTypes;

TYPED_TEST_SUITE(RbfInterpolTypedTestFixture, MyTypes);

TYPED_TEST(RbfInterpolTypedTestFixture, C2Grad) {
  using T = TypeParam;
  rbf_interpol_c2grad_lib<TypeParam>(
      this->p_cell_in.data(), this->rbf_c2grad_idx.data(),
      this->rbf_c2grad_blk.data(), this->rbf_c2grad_coeff.data(),
      this->grad_x.data(), this->grad_y.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->rbf_c2grad_dim, this->nlev, this->nblks_c,
      this->lacc);

  // For each block from i_startblk to i_endblk-1, and for each (i, level)
  // the kernel sums rbf_c2grad_dim contributions, each equal to 1.
  // Therefore, we expect grad_x and grad_y to equal rbf_c2grad_dim (i.e., 10).
  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    for (int jk = 0; jk < this->nlev; ++jk) {
      for (int i = 0; i < this->nproma; ++i) {
        size_t idx = i + static_cast<size_t>(jk) * this->nproma +
                     static_cast<size_t>(jb) * this->nproma * this->nlev;
        EXPECT_NEAR(this->grad_x[idx],
                    static_cast<TypeParam>(this->rbf_c2grad_dim),
                    static_cast<TypeParam>(1e-5))
            << "grad_x failure at block " << jb << ", level " << jk
            << ", index " << i;
        EXPECT_NEAR(this->grad_y[idx],
                    static_cast<TypeParam>(this->rbf_c2grad_dim),
                    static_cast<TypeParam>(1e-5))
            << "grad_y failure at block " << jb << ", level " << jk
            << ", index " << i;
      }
    }
  }
}

TYPED_TEST(RbfInterpolTypedTestFixture, Cell) {
  using T = TypeParam;

  rbf_vec_interpol_cell_lib<T>(
      this->p_vn_in.data(), this->rbf_vec_idx_c.data(),
      this->rbf_vec_blk_c.data(), this->rbf_vec_coeff_c.data(),
      this->p_u_out.data(), this->p_v_out.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->rbf_c2grad_dim, this->nlev, this->nblks_c,
      this->nblks_e, this->lacc, this->acc_async);

  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    for (int jk = 0; jk < this->nlev; ++jk) {
      for (int i = 0; i < this->nproma; ++i) {
        size_t idx = i + static_cast<size_t>(jk) * this->nproma +
                     static_cast<size_t>(jb) * this->nproma * this->nlev;
        EXPECT_NEAR(this->p_u_out[idx], static_cast<T>(this->rbf_vec_dim_c),
                    static_cast<T>(1e-5))
            << "p_u_out failure at block " << jb << ", level " << jk
            << ", index " << i;
      }
    }
  }
}

TYPED_TEST(RbfInterpolTypedTestFixture, Edge) {
  using T = TypeParam;

  rbf_vec_interpol_edge_lib<T>(
      this->p_vn_in.data(), this->rbf_vec_idx_e.data(),
      this->rbf_vec_blk_e.data(), this->rbf_vec_coeff_e.data(),
      this->p_vt_out.data(), this->i_startblk, this->i_endblk,
      this->i_startidx_in, this->i_endidx_in, this->slev, this->elev,
      this->nlev, this->nproma, this->rbf_vec_dim_e, this->nblks_e, this->lacc,
      this->acc_async);

  for (int jb = this->i_startblk; jb < this->i_endblk; ++jb) {
    for (int jk = 0; jk < this->nlev; ++jk) {
      for (int i = 0; i < this->nproma; ++i) {
        size_t idx = i + static_cast<size_t>(jk) * this->nproma +
                     static_cast<size_t>(jb) * this->nproma * this->nlev;
        EXPECT_NEAR(this->p_vt_out[idx], static_cast<T>(this->rbf_vec_dim_e),
                    static_cast<T>(1e-5))
            << "p_vt_out failure at block " << jb << ", level " << jk
            << ", index " << i;
      }
    }
  }
}

// Define a typed test fixture for the functions which have different input and
// output types
template <typename TypePair>
class RbfVecInterpolMixedTestFixture : public ::testing::Test,
                                       public interp_dimensions {
public:
  using InType = typename TypePair::in_type;
  using OutType = typename TypePair::out_type;

  // Constant dimensions.
  static constexpr int nproma = 3;  // inner loop length
  static constexpr int nlev = 4;    // number of vertical levels
  static constexpr int nblks_e = 2; // number of edge blocks (for p_e_in)
  static constexpr int nblks_v =
      2; // number of vertex blocks (for rbf arrays and outputs)
  static constexpr int rbf_vec_dim =
      6; // fixed dimension for rbf vector (stencil points)

  // Parameter values.
  int i_startblk = 0;
  int i_endblk = 1; // Test blocks [0, 1]
  int i_startidx_in = 0;
  int i_endidx_in = nproma - 1; // Full range: 0 .. nproma-1
  int slev = 0;
  int elev = nlev - 1;    // Full vertical range (0 .. nlev-1)
  bool lacc = false;      // Not using ACC-specific behavior.
  bool acc_async = false; // No asynchronous execution.

  // Arrays stored in std::vector.
  std::vector<InType> p_e_in;     // Dimensions: (nproma, nlev, nblks_e)
  std::vector<int> rbf_vec_idx_v; // Dimensions: (rbf_vec_dim, nproma, nblks_v)
  std::vector<int> rbf_vec_blk_v; // Dimensions: (rbf_vec_dim, nproma, nblks_v)
  std::vector<InType>
      rbf_vec_coeff_v;          // Dimensions: (rbf_vec_dim, 2, nproma, nblks_v)
  std::vector<OutType> p_u_out; // Dimensions: (nproma, nlev, nblks_v)
  std::vector<OutType> p_v_out; // Dimensions: (nproma, nlev, nblks_v)

  RbfVecInterpolMixedTestFixture() {
    // Allocate and initialize inputs.
    p_e_in.resize(num_elements_3d<InType>(nproma, nlev, nblks_e),
                  static_cast<InType>(1));
    rbf_vec_idx_v.resize(num_elements_3d<int>(rbf_vec_dim, nproma, nblks_v), 1);
    rbf_vec_blk_v.resize(num_elements_3d<int>(rbf_vec_dim, nproma, nblks_v), 0);
    rbf_vec_coeff_v.resize(
        num_elements_4d<InType>(rbf_vec_dim, 2, nproma, nblks_v),
        static_cast<InType>(1));

    // Allocate output arrays and initialize to zero.
    p_u_out.resize(num_elements_3d<OutType>(nproma, nlev, nblks_v),
                   static_cast<OutType>(0));
    p_v_out.resize(num_elements_3d<OutType>(nproma, nlev, nblks_v),
                   static_cast<OutType>(0));
  }
};

TYPED_TEST_SUITE(RbfVecInterpolMixedTestFixture, MixedTypes);

TYPED_TEST(RbfVecInterpolMixedTestFixture, Vertex) {
  using InType = typename TestFixture::InType;
  using OutType = typename TestFixture::OutType;

  // Call the function with mixed precision.
  rbf_vec_interpol_vertex_lib<InType, OutType>(
      this->p_e_in.data(), this->rbf_vec_idx_v.data(),
      this->rbf_vec_blk_v.data(), this->rbf_vec_coeff_v.data(),
      this->p_u_out.data(), this->p_v_out.data(), this->i_startblk,
      this->i_endblk, this->i_startidx_in, this->i_endidx_in, this->slev,
      this->elev, this->nproma, this->lacc, this->acc_async, this->nlev,
      this->nblks_e, this->nblks_v);

  // Check the outputs only for blocks in the range [i_startblk, i_endblk].
  for (int block = this->i_startblk; block <= this->i_endblk; ++block) {
    for (int level = 0; level < this->nlev; ++level) {
      for (int i = 0; i < this->nproma; ++i) {
        // Compute the linear index for a 3D array in column-major order:
        size_t idx =
            i + level * this->nproma + block * this->nproma * this->nlev;
        // Since every contribution is 1 and there are 6 stencil points,
        // expect 6.
        EXPECT_NEAR(this->p_u_out[idx], static_cast<OutType>(6),
                    static_cast<OutType>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
        EXPECT_NEAR(this->p_v_out[idx], static_cast<OutType>(6),
                    static_cast<OutType>(1e-5))
            << "Failure at block " << block << ", level " << level << ", index "
            << i;
      }
    }
  }
}
